#!/usr/bin/env python
#-*- coding: utf-8 -*-
"""

Utilitaire de mise à jour des variables
pour les versions >= 2.4.1

"""
from creole.upgrade import log
from creole.var_loader import convert_value
from pyeole.i18n import i18n
from tiramisu.setting import owners
from distutils.version import StrictVersion
_ = i18n('creole')

class Upgrade():
    """
    Méthodes pour la mise à niveau des variables
    """
    def __init__(self, config):
        owner = 'upgrade'
        if owner not in dir(owners):
            owners.addowner(owner)
        self.config = config
        self.owner = getattr(owners, owner)
        self.unknown_options = config.impl_get_information(u'unknown_options')

    def get_old_value(self, old_variable, default=None):
        """
        Retourne la valeur d'une variable "disparue"
        """
        try:
            return self.unknown_options[old_variable][u'val']
        except KeyError:
            return default

    def get_value(self, variable, default=None):
        """
        Retourne la valeur d'une variable "connue"
        """
        try:
            path = self.get_path(variable)
        except AttributeError:
            return default
        return self.config.getattr(path,
                                   force_permissive=True)

    def get_path(self, variable):
        """
        Retourne le chemin complet d'une variable
        """
        return self.config.find_first(byname=variable, type_='path')

    def modify_owner(self, path):
        """
        Modifie le propriétaire d'une variable
        """
        option = self.config.unwrap_from_path(path)
        self.config.cfgimpl_get_values().setowner(option,
                                                  self.owner)

    def set_value(self, variable, value):
        """
        Modifie la valeur d'une variable
        """
        path = self.get_path(variable)
        try:
            self.config._setattr(path, value,
                                 force_permissive=True)
            self.modify_owner(path)
            log.info(_(u"Variable updated: {0} = {1}").format(variable, value))
        except ValueError:
            option = self.config.unwrap_from_path(path)
            try:
                # the value could be in Creole 2.3 format #13957
                if not option.impl_is_multi() and type(value) is list and len(value) == 1:
                    value = value[0]
                if value in ['', ['']]:
                    raise ValueError(_(u"empty value"))
                self.config._setattr(path, convert_value(option, value),
                                     force_permissive=True)
                self.modify_owner(path)
                log.info(_(u"Variable updated: {0} = {1}").format(variable, value))
            except Exception, err:
                log.error(_(u"{0} for {1}").format(err, variable))
                self.config.cfgimpl_get_settings()[option].append('load_error')

    def del_value(self, variable):
        path = self.get_path(variable)
        option = self.config.unwrap_from_path(path)
        self.config.cfgimpl_get_values().__delitem__(option)
        log.info(_(u"Variable {0} reinitialized").format(variable))

    def append_value(self, variable, value):
        """
        Ajoute une valeur à une variable multi
        """
        path = self.get_path(variable)
        multi = self.config.getattr(path,
                                    force_permissive=True)
        multi.append(value)
        self.modify_owner(path)

    def modify_last_value(self, variable, value):
        """
        Modifie la dernière valeur d'une variable multi
        """
        path = self.get_path(variable)
        multi = self.config.getattr(path,
                            force_permissive=True)
        multi[-1] = value
        self.modify_owner(path)

    def move(self, old_variable, new_variable):
        """
        Déplace les données d'une variable "disparue"
        vers une nouvelle variable
        """
        if old_variable in self.unknown_options:
            value = self.unknown_options[old_variable][u'val']
            self.set_value(new_variable, value)
            del(self.unknown_options[old_variable])
            log.info(_(u"Variable {0} has been renamed to {1}").format(old_variable, new_variable))

    def copy(self, old_variable, new_variable, only_if_modified=True):
        """
        Copie la valeur d'une variable existante vers une autre
        Si la valeur "old" est une multi et pas la "new" => copie la 1er valeur de la liste
        Si la valeur "old" n'est pas  une multi et la "new" ne l'est pas => transforme la valeur en liste
        only_if_modified: si True ne copie que les valeurs qui sont modifiées
        """
        try:
            # si les deux variables existe => migration
            old_path = self.get_path(old_variable)
            new_path = self.get_path(new_variable)
        except AttributeError:
            pass
        else:
            old_option = self.config.unwrap_from_path(old_path)
            new_option = self.config.unwrap_from_path(new_path)
            # si la nouvelle option n'est pas modifié et si la valeur est modifié ou only_if_modified est False
            if self.config.cfgimpl_get_values().is_default_owner(new_option) and \
                    (not only_if_modified or
                     not self.config.cfgimpl_get_values().is_default_owner(old_option)):
                old_value = self.config.getattr(old_path,
                                                force_permissive=True)
                if old_option.impl_is_multi() and not new_option.impl_is_multi():
                    if len(old_value) != 0:
                        old_value = old_value[0]
                    else:
                        old_value = None
                if not old_option.impl_is_multi() and new_option.impl_is_multi():
                    if old_value is None:
                        old_value = []
                    else:
                        old_value = [old_value]
                self.set_value(new_variable, old_value)


class Upgrade_2_4_1(Upgrade):
    """
    Mise à jour d'une configuration
    de 2.4.0 vers 2.4.1
    """

    def run(self):
        """
        Lancement des traitements
        """
        log.info(_(u"Starting {0} to {1} upgrade").format('2.4.0', '2.4.1'))

        # renommage des variables "era_proxy_bypass"
        for i in range(1, 5):
            self.move('era_proxy_bypass_eth{0}'.format(i), 'proxy_bypass_network_eth{0}'.format(i))

        # fusion des variables "proxy_bypass" et "wpad_exclude"
        if 'adresse_ip_wpad_exclude' in self.unknown_options:
            for idx, value in enumerate(self.get_old_value('adresse_ip_wpad_exclude')):
                interface = self.get_old_value('interface_wpad_exclude')[idx]
                if interface == 'Toutes':
                    interfaces = range(1, 5)
                elif int(interface) in range(1, 5):
                    interfaces = [interface]
                else:
                    log.error(_(u"Invalid value : {0} in old variable {1}").format(interface, 'interface_wpad_exclude'))
                    continue
                for i in interfaces:
                    self.append_value('proxy_bypass_network_eth{0}'.format(i), value)
                    self.modify_last_value('proxy_bypass_netmask_eth{0}'.format(i), self.get_old_value('adresse_netmask_wpad_exclude')[idx])
            del(self.unknown_options['adresse_ip_wpad_exclude'])
            del(self.unknown_options['adresse_netmask_wpad_exclude'])
            del(self.unknown_options['interface_wpad_exclude'])

        # passage à oui des variables "proxy_bypass_ethX" si nécessaire
        for i in range(1, 5):
            if len(self.get_value('proxy_bypass_network_eth{0}'.format(i), [])) > 0:
                self.set_value('proxy_bypass_eth{0}'.format(i), u'oui')

        # transfert des variables nom_domaine_wpad_exclude
        if 'nom_domaine_wpad_exclude' in self.unknown_options:
            for idx, value in enumerate(self.get_old_value('nom_domaine_wpad_exclude')):
                interface = self.get_old_value('nom_interface_wpad_exclude')[idx]
                if interface == 'Toutes':
                    interfaces = range(1, 5)
                elif int(interface) in range(1, 5):
                    interfaces = [interface]
                else:
                    log.error(_(u"Invalid value : {0} in old variable {1}").format(interface, 'nom_interface_wpad_exclude'))
                    continue
                for i in interfaces:
                    self.append_value('proxy_bypass_domain_eth{0}'.format(i), value)
            del(self.unknown_options['nom_domaine_wpad_exclude'])
            del(self.unknown_options['nom_interface_wpad_exclude'])

        # nom_serveur_scribe_dmz/ip_serveur_scribe_dmz => mandatory (#11713)
        if self.get_value('install_scribe_dmz') == u'oui':
            if self.get_value('nom_serveur_scribe_dmz') == None or self.get_value('ip_serveur_scribe_dmz') == None:
                self.set_value('install_scribe_dmz', u'non')


class Upgrade_2_4_2(Upgrade):
    """
    Mise à jour d'une configuration
    de 2.4.1 vers 2.4.2
    """

    def run(self):
        """
        Lancement des traitements
        """
        log.info(_(u"Starting {0} to {1} upgrade").format('2.4.1', '2.4.2'))
        # migration des variables eolesso vers client LDAP #10821
        self.copy('eolesso_port_ldap', 'ldap_port')
        self.copy('eolesso_ldap_reader', 'ldap_reader')
        self.copy('eolesso_ldap_reader_passfile', 'ldap_reader_passfile')
        self.copy('eolesso_ldap_match_attribute', 'ldap_match_attribute')
        self.copy('eolesso_ldap_filter_user', 'ldap_filter_user')
        self.copy('eolesso_ldap_filter_group', 'ldap_filter_group')
        self.copy('eolesso_ldap_dntree_user', 'ldap_dntree_user')
        self.copy('eolesso_ldap_dntree_group', 'ldap_dntree_group')
        self.copy('eolesso_ldap_fill_displayname', 'ldap_fill_displayname')
        self.copy('eolesso_ldap_fill_mail', 'ldap_fill_mail')
        self.copy('eolesso_ldap_fill_fonction', 'ldap_fill_fonction')
        self.copy('eolesso_ldap_fill_categorie', 'ldap_fill_categorie')
        self.copy('eolesso_ldap_fill_rne', 'ldap_fill_rne')
        self.copy('eolesso_ldap_fill_fredurne', 'ldap_fill_fredurne')
        self.copy('eolesso_ldap_fill_displaygroup', 'ldap_fill_displaygroup')

        # migration des variables courier #10987
        courier_val = self.get_old_value('activer_courier')
        if courier_val is not None:
            if courier_val == 'non':
                self.set_value('activer_recuperation_courriel', 'non')
            elif not 'imap' in courier_val:
                self.set_value('activer_courier_imap', 'non')
            if 'pop' in courier_val:
                self.set_value('activer_courier_pop', 'oui')


class Upgrade_2_5_0(Upgrade):
    """
    Mise à jour d'une configuration
    de 2.4.X vers 2.5.0
    """

    def run(self):
        """
        Lancement des traitements
        """
        log.info(_(u"Starting {0} to {1} upgrade").format('2.4.X', '2.5.0'))

        # migration des variables nut #11608
        monitor = self.get_value('nut_monitor_user')
        if monitor != []:
            self.set_value('nut_monitor', 'oui')

        # migration des variables postgresql pour Zéphir #11974
        old_pg_shared_buffers = self.get_value('pg_shared_buffers')
        if old_pg_shared_buffers is not None:
            if int(old_pg_shared_buffers) == 3072:
                self.del_value('pg_shared_buffers')
            else:
                self.set_value('pg_shared_buffers_unit', u'kB')
            self.del_value('pg_effective_cache_size')


class Upgrade_2_5_1(Upgrade):
    """
    Mise à jour d'une configuration
    de 2.5.0 vers 2.5.1
    """

    def run(self):
        """
        Lancement des traitements
        """
        log.info(_(u"Starting {0} to {1} upgrade").format('2.5.0', '2.5.1'))

        # migration des variables zone_forward (#11922)
        zone_forward = self.get_value('nom_zone_forward', [])
        if zone_forward != []:
            self.set_value('activer_zone_forward', 'oui')

        # passage de bacula à bareos (#12425)
        for var in ['activer_bareos_dir', 'activer_bareos_sd',
                    'bareos_dir_name', 'bareos_full_retention',
                    'bareos_full_retention_unit', 'bareos_diff_retention',
                    'bareos_diff_retention_unit', 'bareos_inc_retention',
                    'bareos_inc_retention_unit', 'bareos_max_run_time',
                    'bareos_compression', 'bareos_dir_password',
                    'bareos_fd_password', 'bareos_sd_local',
                    'bareos_sd_adresse', 'bareos_sd_password',
                    'bareos_sd_name', 'bareos_sd_remote_dir_name',
                    'bareos_sd_remote_ip', 'bareos_sd_remote_password']:
            self.move(var.replace('bareos', 'bacula'), var)
        
        if self.get_value('activer_bareos_dir') == u'oui':
            #sauvegarde déjà programmé en sqlite3, ne gère pas la migration vers mysql
            self.set_value('bareos_db_type', 'sqlite3')



def upgrade2(major_version, old_release, current_release, config):
    """
    major_version: version des scripts de migration (ex : 2.4)
    old_release: version du config.eol à migrer (ex : 2.4.0)
    current_release: version du serveur (ex : 2.5.1)
    config: objet de configuration Tiramisu
    """
    def _get_max_release():
        """
        Calcul du dernier numéro de release disponible pour la version majeure
        """
        ends = 0
        for func in globals():
            if func.startswith(func_start):
                ends = max(ends, int(func.split('_')[-1]))
        return ends

    old_version = '.'.join(old_release.split('.')[0:2])
    current_version = '.'.join(current_release.split('.')[0:2])
    func_start = 'Upgrade_' + "_".join(major_version.split('.'))
    if StrictVersion(current_version) == StrictVersion(old_version):
        # upgrade au sein d'une même version
        # ex : 2.5.1 -> 2.5.4 en 2.5
        starts = int(old_release.split('.')[-1])
        ends = int(current_release.split('.')[-1])
    elif StrictVersion(major_version) == StrictVersion(old_version):
        # upgrade "de base" vers une version supérieure
        # ex : 2.4.2 -> 2.6.1 en 2.4
        starts = int(old_release.split('.')[-1])
        ends = _get_max_release()
    elif StrictVersion(major_version) == StrictVersion(current_version):
        # upgrade "final" vers une version supérieure
        # ex : 2.4.2 -> 2.6.1 en 2.6
        starts = -1
        ends = int(current_release.split('.')[-1])
    else:
        # upgrade "intermédiaire" vers une version supérieure
        # ex : 2.4.2 -> 2.6.1 en 2.5
        starts = -1
        ends = _get_max_release()

    for i in xrange(starts + 1, ends + 1):
        func = func_start + '_' + str(i)
        if func in globals():
            upgrade = globals()[func](config)
            upgrade.run()
